'use strict';

var _ = require('../tool/Util');

/**
 * @class pillow.Vector2d Vector2d.
 * @param {Object} options An object literal containing one or more of the following optional properties:
 */

function Vector2d(x, y) {
  this.x = x || 0;
  this.y = y || 0;
}
var proto = {
  initialize: function initialize(x, y) {
    this.x = x;
    this.y = y;
  },
  set: function set(other) {
    this.x = other.x;
    this.y = other.y;
    return this;
  },
  setArray: function setArray(arr) {
    this.x = arr[0];
    this.y = arr[1];
    return this;
  },
  setCoords: function setCoords(x, y) {
    this.x = x;
    this.y = y;
    return this;
  },
  get: function get() {
    return new Vector2d(this.x, this.y);
  },
  mag: function mag() {
    return Math.sqrt(this.x * this.x + this.y * this.y);
  },
  add: function add(other) {
    this.x += other.x;
    this.y += other.y;
    return this;
  },
  addArray: function addArray(arr) {
    this.x += arr[0];
    this.y += arr[1];
    return this;
  },
  addCoords: function addCoords(x, y) {
    this.x += x;
    this.y += y;
    return this;
  },
  sub: function sub(other) {
    this.x -= other.x;
    this.y -= other.y;
    return this;
  },
  subArray: function subArray(arr) {
    this.x -= arr[0];
    this.y -= arr[1];
    return this;
  },
  subCoords: function subCoords(x, y) {
    this.x -= x;
    this.y -= y;
    return this;
  },
  mult: function mult(n) {
    this.x *= n;
    this.y *= n;
    return this;
  },
  scale: function scale(n) {
    this.mult(n);
    return this;
  },
  multVec: function multVec(other) {
    this.x *= other.x;
    this.y *= other.y;
    return this;
  },
  div: function div(n) {
    this.x /= n;
    this.y /= n;
    return this;
  },
  divVec: function divVec(other) {
    this.x /= other.x;
    this.y /= other.y;
    return this;
  },
  dist: function dist(other) {
    var dx = this.x - other.x;
    var dy = this.y - other.y;
    return Math.sqrt(dx * dx + dy * dy);
  },
  dot: function dot(other) {
    return this.x * other.x + this.y * other.y;
  },
  dotCoords: function dotCoords(x, y) {
    return this.x * x + this.y + y;
  },
  normalize: function normalize() {
    var m = this.mag();
    if (m !== 0 && m !== 1) {
      this.div(m);
    }
    return this;
  },
  limit: function limit(max) {
    if (this.mag() > max) {
      this.normalize();
      this.mult(max);
    }
    return this;
  },
  heading2d: function heading2d(x, y) {
    var angle = Math.atan2(-y, x);
    return -1 * angle;
  },
  rotate: function rotate(rads) {
    var s = Math.sin(rads);
    var c = Math.cos(rads);
    var xrot = c * this.x - s * this.y;
    this.y = s * this.x + c * this.y;
    this.x = xrot;
    return this;
  },
  angle: function angle(other) {
    return Math.acos(this.dot(other) / (this.mag() * other.mag()));
  },
  normal: function normal() {
    var temp = this.vector.x;
    this.x = -this.vector.y;
    this.y = temp;
    return this;
  },
  random: function random(mag) {
    this.x = Math.random();
    this.y = Math.random();
    if (mag) this.scale(mag);
    return this;
  },
  zero: function zero() {
    this.x = 0;
    this.y = 0;
  },
  equals: function equals(other) {
    return this.x === other.x && this.y === other.y;
  },
  toString: function toString() {
    return '[' + this.x + ',' + this.y + ']';
  }
};

_.augment(Vector2d, proto);

Vector2d.add = function (one, other) {
  var vec = new Vector2d();
  vec.setCoords(one.x + other.x, one.y + other.y);
  return vec;
};

Vector2d.sub = function (one, other) {
  var vec = new Vector2d();
  vec.setCoords(one.x - other.x, one.y - other.y);
  return vec;
};

Vector2d.dist = function (one, other) {
  var dx = one.x - other.x;
  var dy = one.y - other.y;
  return Math.sqrt(dx * dx + dy * dy);
};

Vector2d.random = function (mag) {
  var vec = new Vector2d(Math.random(), Math.random());
  if (mag) vec.scale(mag);
  return vec;
};

Vector2d.mult = function (one, scalar) {
  var vec = new Vector2d(one.x, one.y);
  vec.x *= scalar;
  vec.y *= scalar;
  return vec;
};

Vector2d.normal = function (vec) {
  return new Vector2d(-vec.y, vec.x);
};

Vector2d.normalize = function (vec) {
  var v = new Vector2d(vec.x, vec.y);
  var m = v.mag();
  if (m !== 0 && m !== 1) {
    v.div(m);
  }
  return v;
};

Vector2d.componentVector = function (vec, directionVec) {
  directionVec.normalize();
  directionVec.mult(vec.dot(directionVec));
  return directionVec;
};

module.exports = Vector2d;